import os
from typing import Any

import numpy as np

import torch
from torch.nn import Module
from torch import Tensor

from ..log import ansi


def _get_device(device: str) -> torch.device:
    match device:
        case "cpu":
            return torch.device("cpu")
        case "cuda":
            return torch.device("cuda")
        case "auto":
            return torch.device("cuda" if torch.cuda.is_available() else "cpu")
        case _:
            raise ValueError("Invalid device")


def _get_dtype(dtype: str) -> torch.dtype:
    match dtype:
        case "qint8":
            return torch.qint8
        case "quint8":
            return torch.quint8
        case "qint32":
            return torch.qint32
        case "float16":
            return torch.float16
        case "float32":
            return torch.float32
        case "float64":
            return torch.float64
        case _:
            raise ValueError("Invalid dtype")


def _compile_model(model: Module, compile_method: str, **kwargs) -> Module:
    def _get_quantization_params() -> tuple[set[type[Module]], torch.dtype]:
        if "layers" not in kwargs or "dtype" not in kwargs:
            raise ValueError("quantization_kwargs must include both 'layers' and 'dtype'")

        return set(kwargs["layers"]), _get_dtype(kwargs["dtype"])

    match compile_method:
        case "none":
            return model

        case "torchscript":
            return torch.jit.script(model)

        case "torchscript_optimized":
            return torch.jit.optimize_for_inference(torch.jit.script(model))

        case "torchscript_quantized":
            import platform
            backend = "qnnpack" if "arm" in platform.processor().lower() else "fbgemm"
            torch.backends.quantized.engine = backend

            layers, dtype = _get_quantization_params()
            return torch.quantization.quantize_dynamic(
                model,
                layers,
                dtype=dtype,
            )

        case "torchscript_quantized_optimized":
            import platform
            backend = "qnnpack" if "arm" in platform.processor().lower() else "fbgemm"
            torch.backends.quantized.engine = backend

            layers, dtype = _get_quantization_params()
            quantized_model = torch.quantization.quantize_dynamic(model, layers, dtype=dtype)
            return torch.jit.optimize_for_inference(torch.jit.script(quantized_model))

        case _:
            raise ValueError("Invalid compile method")


def _load_model(path: str, device: torch.device, dtype: torch.dtype) -> Module:
    if not os.path.exists(path):
        raise FileNotFoundError(f"Model not found at {path}")

    if path.endswith(".pt") or path.endswith(".pth"):
        model = torch.load(path, map_location=device, weights_only=False)
        model.to(device, dtype)
        model.eval()

        print(
            f"{ansi.BOLD}{ansi.GREEN}-> model loaded{ansi.RESET}\n"
            f"   |> path: {path}"
        )

        return model

    if path.endswith(".onnx"):
        try:
            import onnxruntime as ort  # type: ignore
            ort.set_default_logger_severity(3)  # TMP: suppress onnxruntime warnings
        except ImportError:
            print(
                f"{ansi.RED}-> missing library{ansi.RESET}\n"
                f"   |> this method is only available if you have onnxruntime installed\n"
                f"   |> run: pip install onnxruntime"
            )
            raise ImportError("onnxruntime not installed")

        model = ort.InferenceSession(path)

        print(
            f"{ansi.BOLD}{ansi.GREEN}-> model loaded{ansi.RESET}\n"
            f"   |> path: {path}"
        )

        return model

    raise ValueError("Invalid model format")


class Archway:
    """a class for inference with a trained model"""

    def __init__(self,
                 path: str,
                 compile_method: str = "none",
                 device: str = "auto",
                 dtype: str = "float32",
                 quantization_kwargs: dict | None = None,
                 ) -> None:
        """initialize the Archway class

        arguments
        ---------
        path : str
            the path to the model file
        compile_method : str
            the method to compile the model. options: "none", "torchscript", "torchscript_optimized", "torchscript_quantized", "torchscript_quantized_optimized"
        device : str
            the device to run the model on. options: "cpu", "cuda", "auto"
        dtype : str
            the datatype to use for the model. options: "qint8", "quint8", "qint32", "float16", "float32", "float64"
        quantization_kwargs : dict
            keyword arguments for quantization: "layers" (set of layers to quantize), "dtype" (datatype to quantize to)

        raises
        ------
        """
        self.device = _get_device(device)
        self.dtype = _get_dtype(dtype)

        quantization_kwargs = {} if quantization_kwargs is None else quantization_kwargs
        self.model: Module = _load_model(path, self.device, self.dtype)
        self.model = _compile_model(self.model, compile_method, **quantization_kwargs)

        self.predict = self.predict_torch if isinstance(self.model, Module) else self.predict_onnx

    def predict_torch(self, input: np.ndarray | Tensor) -> Any:
        """predict with a PyTorch model

        arguments
        ---------
        input : np.ndarray | Tensor
            the input data

        returns
        -------
        np.ndarray
            the predicted values
        """
        if isinstance(input, np.ndarray):
            input = torch.from_numpy(input).to(self.device, self.dtype)
        elif isinstance(input, Tensor):
            input = input.to(self.device, self.dtype)
        else:
            raise TypeError("input must be a numpy.ndarray or torch.Tensor")

        with torch.inference_mode():
            return self.model(input)

    def predict_onnx(self, input: np.ndarray) -> np.ndarray:
        """predict with an ONNX model

        arguments
        ---------
        input : np.ndarray
            the input data

        returns
        -------
        np.ndarray
            the predicted values
        """
        return self.model.run(None, {"input": input})[0]  # type: ignore

    def serve(self,
              host: str,
              port: int,
              pydantic_model: type,
              endpoint: str = "/predict",
              ) -> None:
        """serve the model with FastAPI

        arguments
        ---------
        host : str
            the host to run the server on
        port : int
            the port to run the server on
        pydantic_model : type
            a Pydantic model for the input data
        endpoint : str
            the name of the endpoint to serve the model on

        raises
        ------
        ImportError
            if the required libraries are not installed: uvicorn, pydantic, fastapi for the server and onnxruntime for ONNX models
        TypeError
            if pydantic_model is not a subclass of pydantic
        """
        try:
            import uvicorn  # type: ignore
            from pydantic import BaseModel  # type: ignore
            from fastapi import FastAPI, HTTPException  # type: ignore
        except ImportError:
            print(
                f"{ansi.RED}-> missing libraries for server{ansi.RESET}\n"
                f"   |> this method is only available if you have uvicorn, pydantic, and fastapi installed\n"
                f"   |> run: pip install uvicorn pydantic fastapi"
            )
            raise ImportError("missing libraries for server")

        if not issubclass(pydantic_model, BaseModel):
            raise TypeError("pydantic_model must be a subclass of pydantic.BaseModel")

        app = FastAPI()

        @app.post(endpoint)
        async def predict(data: pydantic_model) -> dict:  # type: ignore
            try:
                input = np.array([list(data.dict().values())]).astype(np.float32)
                output = self.predict(input)
                return {"output": output.tolist()}

            except Exception as e:
                print(e)
                raise HTTPException(status_code=500, detail=str(e))

        print()
        uvicorn.run(app, host=host, port=port)



from __future__ import annotations

import torch
from torch import Size, Tensor
from torch.distributions import Bernoulli

from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import (
    Kernel,
    LinearKernel,
    MaternKernel,
    PeriodicKernel,
    RBFKernel,
    RFFKernel,
    RQKernel,
    ScaleKernel,
)
from gpytorch.likelihoods import BernoulliLikelihood, Likelihood, _OneDimensionalLikelihood
from gpytorch.means import ConstantMean, Mean, ZeroMean
from gpytorch.models import ApproximateGP
from gpytorch.variational import NaturalVariationalDistribution, VariationalStrategy

KernelSpec = str | type[Kernel] | Kernel
MeanSpec = str | type[Mean] | Mean
LikelihoodSpec = str | type[Likelihood] | Likelihood


def _instantiate_with_fallbacks(class_type: type, kwargs_options: list[dict]) -> object:
    for kwargs in kwargs_options:
        try:
            return class_type(**kwargs)
        except TypeError:
            continue
    raise ValueError(f"could not instantiate class {class_type.__name__} with supported constructor patterns")


def _get_kernel(spec: KernelSpec, n_inputs: int, use_ard: bool, batch_shape: Size) -> Kernel:
    ard_num_dims = n_inputs if use_ard else None

    if isinstance(spec, Kernel):
        return spec

    if isinstance(spec, type):
        if not issubclass(spec, Kernel):
            raise ValueError("kernel class must inherit gpytorch.kernels.Kernel")
        return _instantiate_with_fallbacks(
            spec,
            [
                {"ard_num_dims": ard_num_dims, "batch_shape": batch_shape},
                {"batch_shape": batch_shape},
                {"ard_num_dims": ard_num_dims},
                {},
            ],
        )  # type: ignore[return-value]

    match spec.lower():
        case "rq":
            return RQKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "matern":
            return MaternKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "rbf":
            return RBFKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "rff":
            return RFFKernel(num_samples=n_inputs, batch_shape=batch_shape)
        case "periodic":
            return PeriodicKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "linear":
            return LinearKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case _:
            raise ValueError(f"unknown kernel: {spec}")


def _get_mean_module(spec: MeanSpec, batch_shape: Size) -> Mean:
    if isinstance(spec, Mean):
        return spec

    if isinstance(spec, type):
        if not issubclass(spec, Mean):
            raise ValueError("mean class must inherit gpytorch.means.Mean")
        return _instantiate_with_fallbacks(spec, [{"batch_shape": batch_shape}, {}])  # type: ignore[return-value]

    match spec.lower():
        case "zero":
            return ZeroMean(batch_shape=batch_shape)
        case "constant":
            return ConstantMean(batch_shape=batch_shape)
        case _:
            raise ValueError(f"unknown mean: {spec}")


class PGLikelihood(_OneDimensionalLikelihood):
    def expected_log_prob(self, target: Tensor, input: MultivariateNormal, *args, **kwargs) -> Tensor:
        mean, variance = input.mean, input.variance
        raw_second_moment = variance + mean.pow(2)

        target = target.to(mean.dtype).mul(2.0).sub(1.0)

        c = raw_second_moment.detach().sqrt().clamp_min(1e-8)
        half_omega = 0.25 * torch.tanh(0.5 * c) / c

        res = 0.5 * target * mean - half_omega * raw_second_moment
        return res.sum(dim=-1)

    def forward(self, function_samples: Tensor, *args, **kwargs) -> Bernoulli:
        return Bernoulli(logits=function_samples)

    def marginal(self, function_dist: MultivariateNormal, *args, **kwargs) -> Bernoulli:
        def probability_fn(function_samples: Tensor) -> Tensor:
            return self.forward(function_samples).probs

        probs = self.quadrature(probability_fn, function_dist)
        return Bernoulli(probs=probs)


def _get_likelihood(spec: LikelihoodSpec) -> Likelihood:
    if isinstance(spec, Likelihood):
        return spec

    if isinstance(spec, type):
        if not issubclass(spec, Likelihood):
            raise ValueError("likelihood class must inherit gpytorch.likelihoods.Likelihood")
        return _instantiate_with_fallbacks(spec, [{}])  # type: ignore[return-value]

    match spec.lower():
        case "pg" | "polya_gamma" | "polya-gamma":
            return PGLikelihood()
        case "bernoulli":
            return BernoulliLikelihood()
        case _:
            raise ValueError(f"unknown likelihood: {spec}")


class GPBinaryClassifier(ApproximateGP):
    def __init__(self, *,
                 n_inducing_points: int,
                 n_inputs: int,
                 kernel: KernelSpec = "rq",
                 mean: MeanSpec = "zero",
                 likelihood: LikelihoodSpec = "pg",
                 use_ard: bool = True,
                 ) -> None:
        batch_shape = Size([])
        inducing_points = torch.randn(n_inducing_points, n_inputs, dtype=torch.float32)

        variational_distribution = NaturalVariationalDistribution(n_inducing_points)
        variational_strategy = VariationalStrategy(
            self,
            inducing_points,
            variational_distribution,
            learn_inducing_locations=True,
        )

        super().__init__(variational_strategy)

        self.mean_module = _get_mean_module(mean, batch_shape)
        self.covar_module = ScaleKernel(
            _get_kernel(kernel, n_inputs, use_ard, batch_shape),
            batch_shape=batch_shape,
        )
        self.likelihood = _get_likelihood(likelihood)

    def forward(self, x: Tensor) -> MultivariateNormal:
        mean_x: Tensor = self.mean_module(x)  # type: ignore[assignment]
        covar_x: Tensor = self.covar_module(x)  # type: ignore[assignment]
        return MultivariateNormal(mean_x, covar_x)



from torch import Tensor
from torch.nn import (
    Conv1d,
    Dropout,
    GELU,
    Linear,
    Module,
    ModuleList,
    ReLU,
    Sequential,
    Sigmoid,
    Tanh,
)


def get_activation_function(activation: str) -> Module:
    match activation:
        case 'relu':
            return ReLU()
        case 'sigmoid':
            return Sigmoid()
        case 'tanh':
            return Tanh()
        case 'gelu':
            return GELU()
        case _:
            raise ValueError(f'activation function {activation} not supported')


class Chomp1d(Module):
    def __init__(self, chomp_size: int) -> None:
        super().__init__()
        self.chomp_size = chomp_size

    def forward(self, x: Tensor) -> Tensor:
        if self.chomp_size == 0:
            return x
        return x[:, :, :-self.chomp_size]


class TemporalBlock(Module):
    def __init__(self, *,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 dilation: int,
                 dropout: float,
                 activation: str,
                 ) -> None:
        super().__init__()
        padding = (kernel_size - 1) * dilation

        self.conv1 = Conv1d(
            in_channels,
            out_channels,
            kernel_size,
            stride=1,
            padding=padding,
            dilation=dilation,
        )
        self.chomp1 = Chomp1d(padding)
        self.activation1 = get_activation_function(activation)
        self.dropout1 = Dropout(dropout)

        self.conv2 = Conv1d(
            out_channels,
            out_channels,
            kernel_size,
            stride=1,
            padding=padding,
            dilation=dilation,
        )
        self.chomp2 = Chomp1d(padding)
        self.activation2 = get_activation_function(activation)
        self.dropout2 = Dropout(dropout)

        self.downsample = Conv1d(in_channels, out_channels, kernel_size=1) if in_channels != out_channels else None
        self.out_activation = get_activation_function(activation)

    def forward(self, x: Tensor) -> Tensor:
        out = self.conv1(x)
        out = self.chomp1(out)
        out = self.activation1(out)
        out = self.dropout1(out)

        out = self.conv2(out)
        out = self.chomp2(out)
        out = self.activation2(out)
        out = self.dropout2(out)

        residual = x if self.downsample is None else self.downsample(x)
        return self.out_activation(out + residual)


class TCNRegressor(Module):
    def __init__(self, *,
                 n_inputs: int,
                 n_outputs: int,
                 channel_sizes: list[int],
                 kernel_size: int = 3,
                 feedforward_dims: list[int] = [],
                 dropout: float = 0.0,
                 activation: str = 'relu',
                 ) -> None:
        super().__init__()
        self.n_outputs = n_outputs

        if n_inputs <= 0:
            raise ValueError('n_inputs must be greater than 0')
        if len(channel_sizes) == 0:
            raise ValueError('channel_sizes must contain at least one element')

        self.tcn_blocks = ModuleList()
        in_channels = 1
        for i, out_channels in enumerate(channel_sizes):
            dilation = 2 ** i
            block = TemporalBlock(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                dilation=dilation,
                dropout=dropout,
                activation=activation,
            )
            self.tcn_blocks.append(block)
            in_channels = out_channels

        self.head = ModuleList()
        for i, ff_dim in enumerate(feedforward_dims):
            linear = Linear(
                in_features=channel_sizes[-1] if i == 0 else feedforward_dims[i - 1],
                out_features=ff_dim,
            )
            activation_function = get_activation_function(activation)
            dropout_layer = Dropout(p=dropout)
            self.head.append(linear)
            self.head.append(activation_function)
            self.head.append(dropout_layer)

        final_in_features = channel_sizes[-1] if len(feedforward_dims) == 0 else feedforward_dims[-1]
        self.output_layer = Linear(final_in_features, n_outputs)

    def forward(self, x: Tensor) -> Tensor:
        if x.ndim == 2:
            x = x.unsqueeze(1)
        elif x.ndim != 3:
            raise ValueError('input must have shape [batch, sequence] or [batch, channels, sequence]')

        for block in self.tcn_blocks:
            x = block(x)

        x = x[:, :, -1]

        for layer in self.head:
            x = layer(x)

        x = self.output_layer(x)
        return x.squeeze(-1) if self.n_outputs == 1 else x



from torch import Tensor
from torch.nn import (
    AdaptiveAvgPool1d,
    BatchNorm1d,
    Conv1d,
    Dropout,
    GELU,
    Linear,
    MaxPool1d,
    Module,
    ModuleList,
    ReLU,
    Sequential,
    Sigmoid,
    Tanh,
)


def get_activation_function(activation: str) -> Module:
    match activation:
        case 'relu':
            return ReLU()
        case 'sigmoid':
            return Sigmoid()
        case 'tanh':
            return Tanh()
        case 'gelu':
            return GELU()
        case _:
            raise ValueError(f'activation function {activation} not supported')


class ResidualBlock1D(Module):
    def __init__(self, *,
                 in_channels: int,
                 out_channels: int,
                 stride: int,
                 activation: str,
                 ) -> None:
        super().__init__()

        self.conv1 = Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
        )
        self.bn1 = BatchNorm1d(out_channels)
        self.activation1 = get_activation_function(activation)

        self.conv2 = Conv1d(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )
        self.bn2 = BatchNorm1d(out_channels)
        self.activation2 = get_activation_function(activation)

        if stride != 1 or in_channels != out_channels:
            self.downsample = Sequential(
                Conv1d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False),
                BatchNorm1d(out_channels),
            )
        else:
            self.downsample = None

    def forward(self, x: Tensor) -> Tensor:
        residual = x if self.downsample is None else self.downsample(x)

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.activation1(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = out + residual
        return self.activation2(out)


class ResNet1DRegressor(Module):
    def __init__(self, *,
                 n_inputs: int,
                 n_outputs: int,
                 channels: list[int],
                 blocks_per_stage: int = 2,
                 feedforward_dims: list[int] = [],
                 dropout: float = 0.0,
                 activation: str = 'relu',
                 ) -> None:
        super().__init__()
        self.n_outputs = n_outputs

        if n_inputs <= 0:
            raise ValueError('n_inputs must be greater than 0')
        if len(channels) == 0:
            raise ValueError('channels must contain at least one element')
        if blocks_per_stage <= 0:
            raise ValueError('blocks_per_stage must be greater than 0')

        self.stem = Sequential(
            Conv1d(1, channels[0], kernel_size=7, stride=2, padding=3, bias=False),
            BatchNorm1d(channels[0]),
            get_activation_function(activation),
            MaxPool1d(kernel_size=3, stride=2, padding=1),
        )

        self.stages = ModuleList()
        in_channels = channels[0]
        for stage_idx, out_channels in enumerate(channels):
            for block_idx in range(blocks_per_stage):
                stride = 2 if stage_idx > 0 and block_idx == 0 else 1
                block = ResidualBlock1D(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    stride=stride,
                    activation=activation,
                )
                self.stages.append(block)
                in_channels = out_channels

        self.global_pool = AdaptiveAvgPool1d(1)

        self.head = ModuleList()
        for i, ff_dim in enumerate(feedforward_dims):
            linear = Linear(
                in_features=channels[-1] if i == 0 else feedforward_dims[i - 1],
                out_features=ff_dim,
            )
            activation_function = get_activation_function(activation)
            dropout_layer = Dropout(p=dropout)
            self.head.append(linear)
            self.head.append(activation_function)
            self.head.append(dropout_layer)

        final_in_features = channels[-1] if len(feedforward_dims) == 0 else feedforward_dims[-1]
        self.output_layer = Linear(final_in_features, n_outputs)

    def forward(self, x: Tensor) -> Tensor:
        if x.ndim == 2:
            x = x.unsqueeze(1)
        elif x.ndim != 3:
            raise ValueError('input must have shape [batch, sequence] or [batch, channels, sequence]')

        x = self.stem(x)
        for stage in self.stages:
            x = stage(x)

        x = self.global_pool(x)
        x = x.squeeze(-1)

        for layer in self.head:
            x = layer(x)

        x = self.output_layer(x)
        return x.squeeze(-1) if self.n_outputs == 1 else x



import torch
import numpy as np
import random


def set_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)



import os
import random
import shutil

import pandas as pd


EXTENSIONS = (".jpg", ".jpeg", ".png", ".ppm", ".bmp", ".pgm", ".tif", ".tiff", ".webp", ".csv")


def split_dataset(dataset: str,
                  destination: str,
                  task: str,
                  splits: tuple[float, float, float],
                  shuffle: bool,
                  file_based_split: bool = False,
                  ) -> None:
    """ Split a dataset into training, validation, and testing sets.

    Parameters
    ----------
    dataset : str
        The path to the dataset.
    destination : str
        The path to save the split dataset.
    task : str
        The task type, either "classification" or "regression".
    splits : tuple[float, float, float]
        The split ratios for training, validation, and testing sets.
    shuffle : bool
        Whether to shuffle the dataset before splitting.
    file_based_split : bool, optional
        If set to True, the dataset will be split by dividing each file into the training, validation, and testing sets. If set to False, the dataset will be split by moving entire files into the training, validation, and testing sets (only valid for regression). Defaults to False.

    Raises
    ------
    ValueError
        If the splits do not sum to 1.0, if the training split is 0, or if both the validation and testing splits are 0.
        If the task is not supported.
        If the dataset is not a directory.
    """
    train_split: float = splits[0]
    val_split: float = splits[1]
    test_split: float = splits[2]

    _check_arguments(dataset, task, train_split, val_split, test_split)

    os.makedirs(os.path.join(destination, "training"), exist_ok=True)
    os.makedirs(os.path.join(destination, "validation"), exist_ok=True) if val_split else None
    os.makedirs(os.path.join(destination, "testing"), exist_ok=True) if test_split else None

    if task == "regression":
        split_function = _file_based_regression_split if file_based_split else _dataset_based_regression_split
    elif task == "classification":
        split_function = _classification_split
    else:
        raise ValueError(f"Task {task} not supported")

    split_function(dataset, destination, shuffle, train_split, val_split, test_split)


def _check_arguments(dataset: str,
                     task: str,
                     train_split: float,
                     val_split: float,
                     test_split: float,
                     ) -> None:
    if abs(train_split + val_split + test_split - 1.0) > 1e-6:
        raise ValueError("Splits must sum to 1.0")
    if not train_split:
        raise ValueError("Train split must be greater than 0")
    if not val_split and not test_split:
        raise ValueError("At least one of val split or test split must be greater than 0")
    if task not in ("classification", "regression"):
        raise ValueError(f"Task {task} not supported")
    if not os.path.isdir(dataset):
        raise ValueError("Dataset must be a directory")


def _dataset_based_regression_split(dataset: str,
                                    destination: str,
                                    shuffle: bool,
                                    train_split: float,
                                    val_split: float,
                                    test_split: float,
                                    ) -> None:
    files = os.listdir(dataset)
    files = [file for file in files if file.lower().endswith(".csv")]

    if shuffle:
        random.shuffle(files)

    n_files = len(files)
    train_index = int(train_split * n_files)
    val_index = train_index + int(val_split * n_files)
    test_index = val_index + int(test_split * n_files)

    train = files[:train_index]
    val = files[train_index:val_index] if val_split else []
    test = files[val_index:test_index] if test_split else []

    for file in train:
        shutil.copyfile(os.path.join(dataset, file), os.path.join(destination, "training", file))

    for file in val:
        shutil.copyfile(os.path.join(dataset, file), os.path.join(destination, "validation", file))

    for file in test:
        shutil.copyfile(os.path.join(dataset, file), os.path.join(destination, "testing", file))


def _file_based_regression_split(dataset: str,
                                 destination: str,
                                 shuffle: bool,
                                 train_split: float,
                                 val_split: float,
                                 test_split: float,
                                 ) -> None:
    files = [file for file in os.listdir(dataset) if file.lower().endswith(".csv")]
    for file in files:
        file_path = os.path.join(dataset, file)
        df = pd.read_csv(file_path)
        n_samples = len(df)
        indices = list(range(n_samples))

        if shuffle:
            random.shuffle(indices)

        train_index = int(train_split * n_samples)
        val_index = train_index + int(val_split * n_samples)
        test_index = val_index + int(test_split * n_samples)

        train = df.iloc[indices[:train_index]]
        val = df.iloc[indices[train_index:val_index]] if val_split else pd.DataFrame()
        test = df.iloc[indices[val_index:test_index]] if test_split else pd.DataFrame()

        train_path = os.path.join(destination, "training", file)
        train.to_csv(train_path, index=False)

        val_path = os.path.join(destination, "validation", file)
        val.to_csv(val_path, index=False) if val_split else None

        test_path = os.path.join(destination, "testing", file)
        test.to_csv(test_path, index=False) if test_split else None


def _classification_split(dataset: str,
                          destination: str,
                          shuffle: bool,
                          train_split: float,
                          val_split: float,
                          test_split: float,
                          ) -> None:
    classes = os.listdir(dataset)
    classes = [class_ for class_ in classes if os.path.isdir(os.path.join(dataset, class_))]

    for class_ in classes:
        cls_path = os.path.join(dataset, class_)
        images = [os.path.join(cls_path, img) for img in os.listdir(cls_path)]
        images = [img for img in images if img.lower().endswith(EXTENSIONS)]

        if shuffle:
            random.shuffle(images)

        n_images = len(images)
        train_index = int(train_split * n_images)
        val_index = train_index + int(val_split * n_images)
        test_index = val_index + int(test_split * n_images)

        train = images[:train_index]
        val = images[train_index:val_index] if val_split else []
        test = images[val_index:test_index] if test_split else []

        train_path = os.path.join(destination, "training", class_)
        os.makedirs(train_path, exist_ok=True)
        for img in train:
            train_img = os.path.join(train_path, os.path.basename(img))
            shutil.copyfile(img, train_img)

        val_path = os.path.join(destination, "validation", class_)
        os.makedirs(val_path, exist_ok=True)
        for img in val:
            val_img = os.path.join(val_path, os.path.basename(img))
            shutil.copyfile(img, val_img)

        test_path = os.path.join(destination, "testing", class_)
        os.makedirs(test_path, exist_ok=True)
        for img in test:
            test_img = os.path.join(test_path, os.path.basename(img))
            shutil.copyfile(img, test_img)



import gc
import os
import platform
import time
from typing import Any, Callable

import numpy as np

import torch
from gpytorch.mlls import MarginalLogLikelihood
from torch import Tensor
from torch.distributions import Distribution
from torch.nn import Module
from torch.optim import LBFGS, Optimizer
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.utils.data import DataLoader

from ..dataclasses import EvalInfo, TrainInfo
from ..dataset import CatastaDataset
from ..log import ansi
from .utils import ModelStateManager, TrainingLogger, get_loss_function, get_optimizer

SchedulerType = Any
SchedulerConfig = SchedulerType | tuple[SchedulerType, str]
EarlyStoppingConfig = tuple[str, float | int, float | int] | None


def _get_device(device: str) -> torch.device:
    match device:
        case "cpu":
            return torch.device("cpu")
        case "cuda":
            return torch.device("cuda")
        case "auto":
            return torch.device("cuda" if torch.cuda.is_available() else "cpu")
        case _:
            raise ValueError("Invalid device")


def _normalize_scheduler(scheduler: SchedulerConfig | None) -> tuple[SchedulerType | None, str]:
    if scheduler is None:
        return None, "epoch"

    if isinstance(scheduler, tuple):
        if len(scheduler) != 2 or not isinstance(scheduler[1], str):
            raise ValueError("scheduler tuple must be (scheduler_instance, step_mode)")
        return scheduler[0], scheduler[1]

    return scheduler, "epoch"


class Scaffold:
    def __init__(
        self,
        *,
        model: Module,
        dataset: CatastaDataset,
        optimizer: str | Optimizer,
        loss_function: str | Module,
        device: str = "auto",
        verbose: bool = True,
    ) -> None:
        self.task: str = dataset.task

        self.device: torch.device = _get_device(device)
        self.input_dtype: torch.dtype = torch.float32
        self.output_dtype: torch.dtype = torch.float32 if self.task == "regression" else torch.long

        self.verbose: bool = verbose
        self.model: Module = model.to(self.device, self.input_dtype)
        self.dataset: CatastaDataset = dataset
        self.optimizer_id: str | Optimizer = optimizer
        self.loss_function_id: str | Module = loss_function

        if not self.verbose:
            return

        n_params = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        device_name = torch.cuda.get_device_name() if self.device.type == "cuda" else platform.processor()
        n_samples = len(self.dataset.train)  # type: ignore[arg-type]
        print(
            f"{ansi.BOLD}{ansi.GREEN}-> training info\n{ansi.RESET}"
            f"   |> task:    {self.task}\n"
            f"   |> dataset: {self.dataset.root} ({n_samples} samples)\n"
            f"   |> model:   {self.model.__class__.__name__} ({n_params} parameters)\n"
            f"   |> device:  {self.device} ({device_name})"
        )

    def train(
        self,
        *,
        epochs: int,
        batch_size: int,
        lr: float,
        scheduler: SchedulerConfig | None = None,
        early_stopping: EarlyStoppingConfig = None,
        shuffle: bool = True,
        data_loader_workers: int = 0,
    ) -> TrainInfo:
        optimizer: Optimizer = get_optimizer(self.optimizer_id, self.model, lr)
        loss_function: Module = get_loss_function(self.loss_function_id, self.model, len(self.dataset.train))  # type: ignore[arg-type]

        scheduler_obj, scheduler_step_mode = _normalize_scheduler(scheduler)
        if scheduler_step_mode == "val_loss" and not isinstance(scheduler_obj, ReduceLROnPlateau):
            raise ValueError("scheduler mode 'val_loss' requires torch.optim.lr_scheduler.ReduceLROnPlateau")

        model_state_manager = ModelStateManager("none" if early_stopping is None else early_stopping)
        training_logger = TrainingLogger(self.task, epochs)

        train_data_loader = DataLoader(
            self.dataset.train,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=data_loader_workers,
            pin_memory=self.device.type == "cuda",
        )
        val_data_loader = DataLoader(
            self.dataset.validation,
            batch_size=batch_size,
            shuffle=False,
            num_workers=data_loader_workers,
            pin_memory=self.device.type == "cuda",
        )

        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        for epoch in range(epochs):
            try:
                start_time = time.time()

                self.model.train()
                train_loss = self._epoch(train_data_loader, loss_function, optimizer)

                self.model.eval()
                with torch.no_grad():
                    val_loss = self._epoch(val_data_loader, loss_function, None)

                model_state_manager(self.model, val_loss)

                training_logger.log(
                    train_loss=train_loss,
                    val_loss=val_loss,
                    lr=float(optimizer.param_groups[0]["lr"]),
                    epoch=epoch + 1,
                    time_per_epoch=time.time() - start_time,
                )

                self._step_scheduler(scheduler_obj, scheduler_step_mode, train_loss, val_loss)

                if self.verbose:
                    print(training_logger)

                if model_state_manager.stop:
                    if self.verbose:
                        print(
                            f"\n{ansi.BOLD}{ansi.YELLOW}-> early stopping{ansi.RESET}\n"
                            f"   |> epoch: {epoch + 1}"
                        )
                    break

            except KeyboardInterrupt:
                if self.verbose:
                    print(
                        f"\n{ansi.BOLD}{ansi.YELLOW}-> training interrupted{ansi.RESET}\n"
                        f"   |> epoch: {epoch + 1}"
                    )
                break

        train_info = training_logger.get_info()
        model_state_manager.load_best_model_state(self.model)
        return train_info

    def _step_scheduler(
        self,
        scheduler: SchedulerType | None,
        step_mode: str,
        train_loss: float,
        val_loss: float,
    ) -> None:
        if scheduler is None:
            return

        if step_mode == "epoch":
            scheduler.step()
            return

        if step_mode == "train_loss":
            scheduler.step(train_loss)
            return

        if step_mode == "val_loss":
            scheduler.step(val_loss)
            return

        raise ValueError("scheduler step mode must be one of: 'epoch', 'train_loss', 'val_loss'")

    def _epoch(
        self,
        data_loader: DataLoader,
        loss_function: Module,
        optimizer: Optimizer | None,
    ) -> float:
        cumulated_loss = torch.tensor(0.0, device=self.device)
        total_samples = 0

        for inputs, targets in data_loader:
            inputs = inputs.to(self.device, self.input_dtype, non_blocking=True)
            targets = targets.to(self.device, self.output_dtype, non_blocking=True)

            if optimizer is not None and isinstance(optimizer, LBFGS):
                def closure() -> Tensor:
                    optimizer.zero_grad(set_to_none=True)
                    output = self.model(inputs)
                    loss = loss_function(output, targets)  # type: ignore[arg-type]
                    if isinstance(loss_function, MarginalLogLikelihood):
                        loss = -loss
                    loss.backward()
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)  # type: ignore[arg-type]
                    return loss

                loss = optimizer.step(closure)
                batch_loss = loss.detach()
            else:
                if optimizer is not None:
                    optimizer.zero_grad(set_to_none=True)

                output = self.model(inputs)
                loss = loss_function(output, targets)  # type: ignore[arg-type]
                if isinstance(loss_function, MarginalLogLikelihood):
                    loss = -loss

                if optimizer is not None:
                    loss.backward()
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)  # type: ignore[arg-type]
                    optimizer.step()

                batch_loss = loss.detach()

            cumulated_loss += batch_loss * inputs.shape[0]
            total_samples += inputs.shape[0]

        if total_samples == 0:
            return 0.0

        return float((cumulated_loss / total_samples).item())

    @torch.no_grad()
    def evaluate(
        self,
        batch_size: int | None = None,
        prediction_function: Callable[[Module, Any], Any] | None = None,
    ) -> EvalInfo:
        self.model.eval()

        effective_batch_size = len(self.dataset.test) if batch_size is None else batch_size  # type: ignore[arg-type]
        data_loader = DataLoader(
            self.dataset.test,
            batch_size=effective_batch_size,
            shuffle=False,
            num_workers=0,
            pin_memory=self.device.type == "cuda",
        )

        true_input: list[np.ndarray] = []
        true_output: list[np.ndarray] = []
        predicted_output: list[np.ndarray] = []
        for x, y in data_loader:
            x = x.to(self.device, self.input_dtype)
            output = self.model(x) if prediction_function is None else prediction_function(self.model, x)

            if isinstance(output, Distribution):
                output = output.mean

            if isinstance(output, Tensor):
                output_np = output.detach().cpu().numpy()
            else:
                output_np = np.asarray(output)

            true_input.append(x.detach().cpu().numpy())
            true_output.append(y.detach().cpu().numpy())
            predicted_output.append(output_np)

        return EvalInfo(
            task=self.task,
            true_input=np.concatenate(true_input),
            true_output=np.concatenate(true_output),
            predicted_output=np.concatenate(predicted_output),
        )

    def save(self, path: str) -> None:
        is_model_file = path.endswith((".pt", ".pth", ".onnx"))
        if is_model_file:
            parent_dir = os.path.dirname(path)
            if parent_dir and not os.path.exists(parent_dir):
                os.makedirs(parent_dir)
                print(
                    f"{ansi.BOLD}{ansi.YELLOW}-> specified path does not exist{ansi.RESET}\n"
                    f"   |> created directory: {parent_dir}"
                )
        elif not os.path.exists(path):
            os.makedirs(path)
            print(
                f"{ansi.BOLD}{ansi.YELLOW}-> specified path does not exist{ansi.RESET}\n"
                f"   |> created directory: {path}"
            )

        self.model.to("cpu")

        if os.path.isdir(path):
            path = os.path.join(path, f"{self.model.__class__.__name__}.pt")

        if path.endswith(".pt") or path.endswith(".pth"):
            torch.save(self.model, path)
        elif path.endswith(".onnx"):
            try:
                import onnx  # type: ignore # noqa: F401
            except ImportError:
                print(
                    f"{ansi.BOLD}{ansi.RED}-> onnx library not found{ansi.RESET}\n"
                    f"   |> onnx is a special use case, please install it manually\n"
                    f"   |> the model will be saved as a `.pt` file\n"
                    f"   |> run: pip install onnx"
                )
                fallback_path = path.replace(".onnx", ".pt")
                torch.save(self.model, fallback_path)
                print(
                    f"{ansi.BOLD}{ansi.GREEN}-> model saved{ansi.RESET}\n"
                    f"   |> path: {fallback_path}"
                )
                return

            input_shape = self.dataset.train[0][0].shape
            dummy_input = torch.randn(1, *input_shape, dtype=self.input_dtype)
            torch.onnx.export(
                self.model,
                dummy_input,  # type: ignore[arg-type]
                path,
                export_params=True,
                opset_version=11,
                do_constant_folding=True,
                input_names=["input"],
                output_names=["output"],
                dynamic_axes={"input": {0: "batch_size"}, "output": {0: "batch_size"}},
            )
        else:
            raise ValueError("Invalid path. Must be a directory, a `.pt` or `.onnx` file.")

        print(
            f"{ansi.BOLD}{ansi.GREEN}-> model saved{ansi.RESET}\n"
            f"   |> path: {path}"
        )



from torch import Tensor
from torch.nn import (
    Dropout,
    GELU,
    LSTM,
    Linear,
    Module,
    ModuleList,
    ReLU,
    Sigmoid,
    Tanh,
)


def get_activation_function(activation: str) -> Module:
    match activation:
        case 'relu':
            return ReLU()
        case 'sigmoid':
            return Sigmoid()
        case 'tanh':
            return Tanh()
        case 'gelu':
            return GELU()
        case _:
            raise ValueError(f'activation function {activation} not supported')


class LSTMRegressor(Module):
    def __init__(self, *,
                 n_inputs: int,
                 n_outputs: int,
                 hidden_dim: int,
                 n_layers: int = 1,
                 bidirectional: bool = False,
                 feedforward_dims: list[int] = [],
                 dropout: float = 0.0,
                 activation: str = 'relu',
                 ) -> None:
        super().__init__()

        if n_layers <= 0:
            raise ValueError('n_layers must be greater than 0')

        self.n_outputs = n_outputs
        self.lstm = LSTM(
            input_size=1,
            hidden_size=hidden_dim,
            num_layers=n_layers,
            batch_first=True,
            bidirectional=bidirectional,
            dropout=dropout if n_layers > 1 else 0.0,
        )

        recurrent_output_dim = hidden_dim * (2 if bidirectional else 1)
        self.dense_layers = ModuleList()
        for i, ff_dim in enumerate(feedforward_dims):
            linear = Linear(
                in_features=recurrent_output_dim if i == 0 else feedforward_dims[i - 1],
                out_features=ff_dim,
            )
            activation_function = get_activation_function(activation)
            dropout_layer = Dropout(p=dropout)
            self.dense_layers.append(linear)
            self.dense_layers.append(activation_function)
            self.dense_layers.append(dropout_layer)

        final_in_features = recurrent_output_dim if len(feedforward_dims) == 0 else feedforward_dims[-1]
        self.output_layer = Linear(final_in_features, n_outputs)

    def forward(self, x: Tensor) -> Tensor:
        if x.ndim == 2:
            x = x.unsqueeze(-1)
        elif x.ndim != 3:
            raise ValueError('input must have shape [batch, sequence] or [batch, sequence, features]')

        x, _ = self.lstm(x)
        x = x[:, -1, :]

        for layer in self.dense_layers:
            x = layer(x)

        x = self.output_layer(x)
        return x.squeeze(-1) if self.n_outputs == 1 else x



from typing import Callable
from datetime import datetime

import numpy as np

import optuna
from optuna import Trial, Study
from optuna.trial import FrozenTrial

from ..dataclasses import OptimizationInfo, TrialInfo
from .utils import OptimizationLogger, get_sampler
from ..log import ansi


class Foundation:
    def __init__(self, *,
                 hyperparameter_space: dict[str, tuple],
                 objective_function: Callable[[dict, dict], float],
                 sampler: str,
                 n_trials: int,
                 direction: str,
                 objective_extra_parameters: dict | None = None,
                 use_secretary: bool = False,
                 catch_exceptions: bool = False,
                 verbose: bool = True,
                 ) -> None:
        """Initialize the Foundation class.

        Arguments
        ---------
        hyperparameter_space : dict[str, tuple]
            Dictionary containing the hyperparameters to optimize. The keys are the hyperparameter names and the values are tuples containing the hyperparameters search space.
        objective_function : Callable[[dict, dict], float]
            Objective function to optimize. The function should contain two dictionaries as parameters: the first one contains the parameters being optimized, and the second one the `objective_extra_parameters` dictionary.
        objective_extra_parameters : dict
            the extra parameters to include every function call to the objective function
        sampler : str
            Name of the sampler to use. Valid values are "random", "rs", "tpe", "gp", and "bogp".
        n_trials : int
            Number of trials to run.
        direction : str
            Direction of the optimization. Valid values are "minimize" and "maximize".
        use_secretary : bool, optional
            Whether to use the secretary algorithm. The secretary algorithm will reject the first n_trials/e trials and select the best trial after that. Default is False.
        catch_exceptions : bool, optional
            Whether to catch exceptions raised by the objective function. Default is False.
        verbose : bool, optional
            Whether to print optimization information. Default is True.

        Raises
        ------
        ValueError
            If the direction is invalid.
            If the sampler is invalid.
        """
        self.hyperparameter_space = hyperparameter_space
        self.objective_function = objective_function
        self.objective_extra_parameters = {} if objective_extra_parameters is None else objective_extra_parameters
        self.sampler = get_sampler(sampler)
        self.n_trials = n_trials
        self.direction = direction
        self.use_secretary = use_secretary
        self.catch_exceptions = catch_exceptions

        self.verbose = verbose

        if direction not in ["minimize", "maximize"]:
            raise ValueError("Invalid direction")

        if not self.verbose:
            return

        # log optimization info
        n_hyperparameters = len(self.hyperparameter_space)
        n_combinations = np.prod([len(values) for values in self.hyperparameter_space.values()])
        print(
            f"{ansi.BOLD}{ansi.GREEN}-> optimization info{ansi.RESET}\n"
            f"   |> hyperparameters: {n_hyperparameters} ({n_combinations} combinations)\n"
            f"   |> sampler:         {self.sampler.__class__.__name__}\n"
            f"   |> direction:       {self.direction}"
        )

    def optimize(self) -> OptimizationInfo:
        """Optimize the hyperparameters.

        Returns
        -------
        OptimizationInfo
            Object containing the information about the optimization.

        Raises
        ------
        Exception
            If the objective function raises an exception and catch_exceptions is False.
        """
        def optimize_optuna(trial: Trial) -> float:
            params = {name: trial.suggest_categorical(name, values) for name, values in self.hyperparameter_space.items()}

            try:
                metric = self.objective_function(params, self.objective_extra_parameters.copy())
            except Exception as e:
                if isinstance(e, KeyboardInterrupt):
                    print(
                        f"{ansi.BOLD}{ansi.YELLOW}-> optimization interrupted\n"
                        f"   |> trial: {trial.number + 1}{ansi.RESET}"
                    )
                    raise e

                if self.catch_exceptions:
                    print(
                        f"{ansi.BOLD}{ansi.RED}-> exception caught\n"
                        f"   |> trial: {trial.number + 1}\n"
                        f"   |> error: {e}{ansi.RESET}"
                    )
                    metric = np.inf if self.direction == "minimize" else -np.inf
                else:
                    raise e

            return metric

        optimization_logger = OptimizationLogger(self.n_trials)

        def on_trial_complete(study: Study, trial: FrozenTrial) -> None:
            if self.use_secretary and trial.number > self.n_trials / np.e and study.best_trial.number == trial.number:
                study.stop()

            optimization_logger.log(
                trial_number=trial.number + 1,
                metric=trial.value if trial.value is not None else 0.0,  # TMP
                best_trial_number=study.best_trial.number + 1,
                best_metric=study.best_value,
                time_elapsed=trial.duration.total_seconds() if trial.duration is not None else 0.0,  # TMP
            )

            print(optimization_logger) if self.verbose else None

        study = optuna.create_study(
            sampler=self.sampler,
            direction=self.direction,
        )
        study.optimize(
            func=optimize_optuna,
            n_trials=self.n_trials,
            callbacks=[on_trial_complete],
        )

        first_datetime = datetime.now()
        for trial in study.trials:
            if trial.datetime_start is not None:
                first_datetime = trial.datetime_start
                break

        last_datetime = datetime.now()
        for trial in reversed(study.trials):
            if trial.datetime_complete is not None:
                last_datetime = trial.datetime_complete
                break
            if trial.datetime_start is not None:
                last_datetime = trial.datetime_start
                break

        history = []
        for trial in study.trials:
            metric = np.inf if self.direction == "minimize" else -np.inf
            if trial.value is not None:
                metric = trial.value

            trial_start = trial.datetime_start if trial.datetime_start is not None else datetime.now()
            trial_end = trial.datetime_complete if trial.datetime_complete is not None else trial_start
            trial_duration = 0.0
            if trial.duration is not None:
                trial_duration = max(float(trial.duration.total_seconds()), 0.0)

            history.append(TrialInfo(
                number=trial.number + 1,
                hyperparameters=trial.params,
                metric=metric,
                datetime_start=trial_start,
                datetime_end=trial_end,
                duration=trial_duration,
            ))

        duration = max(last_datetime.timestamp() - first_datetime.timestamp(), 0.0)
        return OptimizationInfo(
            best_hyperparameters=study.best_params,
            best_metric=study.best_value,
            best_trial_number=study.best_trial.number + 1,
            datetime_start=first_datetime,
            datetime_end=last_datetime,
            duration=duration,
            history=history,
        )



import torch
from torch import Tensor
from torch.nn import (
    Module,
    Linear,
    LayerNorm,
    Sequential,
    ModuleList,
    Dropout,
    Softmax,
    GELU,
    Identity,
)

from einops import rearrange, reduce
from einops.layers.torch import Rearrange


def _validate_n_patches(n_inputs: int, n_patches: int) -> int:
    if n_patches <= 0:
        raise ValueError("n_patches must be greater than 0")
    if n_patches > n_inputs:
        raise ValueError("n_patches cannot be greater than n_inputs")
    if n_inputs % n_patches != 0:
        raise ValueError(f"sequence length {n_inputs} must be divisible by n_patches {n_patches}")

    return n_inputs // n_patches


def posemb_sincos_1d(patches: Tensor, temperature: int = 10000) -> Tensor:
    n: int = patches.shape[1]
    d_model: int = patches.shape[2]
    device: torch.device = patches.device
    dtype: torch.dtype = patches.dtype

    if (d_model % 2) != 0:
        raise ValueError(f'feature dimension must be multiple of 2 for sincos emb. got {d_model}')

    n_tensor: Tensor = torch.arange(n, device=device)
    omega: Tensor = torch.arange(d_model // 2, device=device) / (d_model // 2 - 1)
    omega = 1.0 / (temperature ** omega)

    n_tensor = n_tensor.flatten()[:, None] * omega[None, :]
    pe: Tensor = torch.cat((n_tensor.sin(), n_tensor.cos()), dim=1)

    return pe.to(dtype)


class FeedForward(Module):
    def __init__(self, *,
                 d_model: int,
                 hidden_dim: int,
                 dropout: float,
                 layer_norm: bool,
                 ) -> None:
        super().__init__()

        self.net = Sequential(
            LayerNorm(d_model) if layer_norm else Identity(),
            Linear(d_model, hidden_dim),
            GELU(),
            Dropout(dropout),
            Linear(hidden_dim, d_model),
            Dropout(dropout),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.net(x)


class Attention(Module):
    def __init__(self, *,
                 d_model: int,
                 n_heads: int,
                 head_dim: int,
                 dropout: float,
                 layer_norm: bool,
                 ) -> None:
        super().__init__()

        self.n_heads: int = n_heads
        self.scale: int = head_dim ** -0.5
        self.norm = LayerNorm(d_model) if layer_norm else Identity()
        self.attend = Softmax(dim=-1)
        self.dropout = Dropout(dropout)

        inner_dim: int = head_dim * n_heads
        self.to_qkv = Linear(d_model, inner_dim * 3, bias=False)
        self.to_out = Sequential(
            Linear(inner_dim, d_model, bias=False),
            Dropout(dropout),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.norm(x)

        qkv: Tensor = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.n_heads), qkv)

        dots: Tensor = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn: Tensor = self.attend(dots)
        attn = self.dropout(attn)

        out: Tensor = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')

        return self.to_out(out)


class Transformer(Module):
    def __init__(self, *,
                 d_model: int,
                 n_layers: int,
                 n_heads: int,
                 head_dim: int,
                 feedforward_dim: int,
                 dropout: float,
                 layer_norm: bool,
                 ) -> None:
        super().__init__()

        self.norm = LayerNorm(d_model) if layer_norm else Identity()
        self.layers = ModuleList([])
        for _ in range(n_layers):
            self.layers.append(ModuleList([
                Attention(d_model=d_model,
                          n_heads=n_heads,
                          head_dim=head_dim,
                          dropout=dropout,
                          layer_norm=layer_norm,
                          ),
                FeedForward(d_model=d_model,
                            hidden_dim=feedforward_dim,
                            dropout=dropout,
                            layer_norm=layer_norm,
                            )
            ]))

    def forward(self, x: Tensor) -> Tensor:
        for attn, ff in self.layers:  # type: ignore
            x = attn(x) + x
            x = ff(x) + x

        x = self.norm(x)

        return x


class TransformerRegressor(Module):
    def __init__(self, *,
                 n_inputs: int,
                 n_outputs: int,
                 n_patches: int,
                 d_model: int,
                 n_layers: int,
                 n_heads: int,
                 feedforward_dim: int,
                 head_dim: int,
                 pooling: str = "mean",
                 dropout: float = 0.0,
                 layer_norm: bool = False,
                 ) -> None:
        super().__init__()
        self.n_outputs = n_outputs
        patch_size: int = _validate_n_patches(n_inputs, n_patches)

        self.to_patch_embedding = Sequential(
            Rearrange('b (n p) -> b n p', p=patch_size),
            LayerNorm(patch_size) if layer_norm else Identity(),
            Linear(patch_size, d_model),
            LayerNorm(d_model) if layer_norm else Identity(),
        )

        self.pos_embedding = posemb_sincos_1d
        self.dropout = Dropout(dropout)
        self.pooling = pooling

        self.transformer = Transformer(
            d_model=d_model,
            n_layers=n_layers,
            n_heads=n_heads,
            head_dim=head_dim,
            feedforward_dim=feedforward_dim,
            dropout=dropout,
            layer_norm=layer_norm,
        )

        linear_head_input_dim: int = d_model * n_patches if pooling == "concat" else d_model

        self.linear_head = Sequential(
            LayerNorm(linear_head_input_dim) if layer_norm else Identity(),
            Linear(linear_head_input_dim, n_outputs),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.to_patch_embedding(x)
        x += self.pos_embedding(x)

        x = self.transformer(x)
        x = self.dropout(x)

        x = reduce(x, 'b n d -> b d', self.pooling) if self.pooling != "concat" else rearrange(x, 'b n d -> b (n d)')
        y = self.linear_head(x)
        return y.squeeze(-1) if self.n_outputs == 1 else y



from __future__ import annotations

from typing import TypeAlias

import torch
from torch import Size, Tensor
from torch.nn import Module

from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import (
    Kernel,
    LinearKernel,
    MaternKernel,
    PeriodicKernel,
    RBFKernel,
    RFFKernel,
    RQKernel,
    ScaleKernel,
)
from gpytorch.likelihoods import (
    BernoulliLikelihood,
    BetaLikelihood,
    GaussianLikelihood,
    LaplaceLikelihood,
    Likelihood,
    MultitaskGaussianLikelihood,
    SoftmaxLikelihood,
    StudentTLikelihood,
)
from gpytorch.means import ConstantMean, Mean, ZeroMean
from gpytorch.models import ApproximateGP
from gpytorch.variational import (
    CholeskyVariationalDistribution,
    IndependentMultitaskVariationalStrategy,
    VariationalStrategy,
)

KernelSpec: TypeAlias = str | type[Kernel] | Kernel
MeanSpec: TypeAlias = str | type[Mean] | Mean
LikelihoodSpec: TypeAlias = str | type[Likelihood] | Likelihood


def _instantiate_with_fallbacks(class_type: type, kwargs_options: list[dict]) -> object:
    for kwargs in kwargs_options:
        try:
            return class_type(**kwargs)
        except TypeError:
            continue
    raise ValueError(f"could not instantiate class {class_type.__name__} with supported constructor patterns")


def _get_kernel(spec: KernelSpec, n_inputs: int, use_ard: bool, batch_shape: Size) -> Kernel:
    ard_num_dims = n_inputs if use_ard else None

    if isinstance(spec, Kernel):
        return spec

    if isinstance(spec, type):
        if not issubclass(spec, Kernel):
            raise ValueError("kernel class must inherit gpytorch.kernels.Kernel")
        return _instantiate_with_fallbacks(
            spec,
            [
                {"ard_num_dims": ard_num_dims, "batch_shape": batch_shape},
                {"batch_shape": batch_shape},
                {"ard_num_dims": ard_num_dims},
                {},
            ],
        )  # type: ignore[return-value]

    match spec.lower():
        case "rq":
            return RQKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "matern":
            return MaternKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "rbf":
            return RBFKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "rff":
            return RFFKernel(num_samples=n_inputs, batch_shape=batch_shape)
        case "periodic":
            return PeriodicKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "linear":
            return LinearKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case _:
            raise ValueError(f"Unknown kernel: {spec}")


def _get_mean_module(spec: MeanSpec, batch_shape: Size) -> Mean:
    if isinstance(spec, Mean):
        return spec

    if isinstance(spec, type):
        if not issubclass(spec, Mean):
            raise ValueError("mean class must inherit gpytorch.means.Mean")
        return _instantiate_with_fallbacks(spec, [{"batch_shape": batch_shape}, {}])  # type: ignore[return-value]

    match spec.lower():
        case "constant":
            return ConstantMean(batch_shape=batch_shape)
        case "zero":
            return ZeroMean(batch_shape=batch_shape)
        case _:
            raise ValueError(f"Unknown mean: {spec}")


def _get_likelihood(spec: LikelihoodSpec, n_outputs: int) -> Likelihood:
    if isinstance(spec, Likelihood):
        return spec

    if isinstance(spec, type):
        if not issubclass(spec, Likelihood):
            raise ValueError("likelihood class must inherit gpytorch.likelihoods.Likelihood")
        if n_outputs > 1:
            return _instantiate_with_fallbacks(spec, [{"num_tasks": n_outputs}, {}])  # type: ignore[return-value]
        return _instantiate_with_fallbacks(spec, [{}])  # type: ignore[return-value]

    if n_outputs > 1:
        return MultitaskGaussianLikelihood(num_tasks=n_outputs)

    match spec.lower():
        case "gaussian":
            return GaussianLikelihood()
        case "bernoulli":
            return BernoulliLikelihood()
        case "laplace":
            return LaplaceLikelihood()
        case "softmax":
            return SoftmaxLikelihood()
        case "studentt":
            return StudentTLikelihood()
        case "beta":
            return BetaLikelihood()
        case _:
            raise ValueError(f"Unknown likelihood: {spec}")


class GPHeadRegressor(ApproximateGP):
    def __init__(
        self,
        *,
        pre_model: Module,
        pre_model_output_dim: int,
        n_inputs: int,
        n_outputs: int,
        n_inducing_points: int,
        kernel: KernelSpec = "rq",
        mean: MeanSpec = "constant",
        likelihood: LikelihoodSpec = "gaussian",
        use_ard: bool = True,
    ) -> None:
        self._n_outputs = n_outputs

        batch_shape = Size([])
        inducing_points = torch.randn(n_inducing_points, n_inputs, dtype=torch.float32)

        variational_distribution = CholeskyVariationalDistribution(
            n_inducing_points,
            batch_shape=batch_shape,
        )
        base_variational_strategy = VariationalStrategy(
            self,
            inducing_points,
            variational_distribution,
            learn_inducing_locations=True,
        )
        variational_strategy = (
            base_variational_strategy
            if n_outputs == 1
            else IndependentMultitaskVariationalStrategy(
                base_variational_strategy=base_variational_strategy,
                num_tasks=n_outputs,
            )
        )

        super().__init__(variational_strategy)
        self.mean_module = _get_mean_module(mean, batch_shape)
        self.covar_module = ScaleKernel(
            _get_kernel(kernel, pre_model_output_dim, use_ard, batch_shape),
            batch_shape=batch_shape,
        )
        self.likelihood = _get_likelihood(likelihood, n_outputs)
        self.pre_model = pre_model

    def forward(self, x: Tensor) -> MultivariateNormal:
        x = self.pre_model(x)
        if x.dim() > 2:
            x = x.view(x.shape[0], -1)
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)  # type: ignore[arg-type]



from typing import NamedTuple
from datetime import datetime

import pandas as pd

from .trial_info import TrialInfo
from ..log import ansi


class OptimizationInfo(NamedTuple):
    """Dataclass containing the optimization information.

    Attributes
    ----------
    best_hyperparameters : dict
        Best hyperparameters found.
    best_metric : float
        Best metric found.
    best_trial_number : int
        Trial number of the best trial.
    duration : float
        Time elapsed during the optimization.
    history : None
        History of the trials.

    Methods
    -------
    to_pandas()
        Convert the OptimizationInfo to a pandas DataFrame.
    """
    best_hyperparameters: dict
    best_metric: float
    best_trial_number: int
    datetime_start: datetime
    datetime_end: datetime
    duration: float
    history: list[TrialInfo]

    def __repr__(self) -> str:
        msg = f"\n{ansi.BOLD}{ansi.BLUE}-> optimization results{ansi.RESET}\n"
        msg += f"   |> trial: {self.best_trial_number}\n"
        msg += f"   |> metric: {self.best_metric:.4f}\n"
        msg += f"   |> hyperparameters:\n"
        for key, value in self.best_hyperparameters.items():
            msg += f"      * {key}: {value}\n"

        return msg.rstrip()

    def to_pandas(self) -> pd.DataFrame:
        """Convert the OptimizationInfo to a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the optimization information.
        """
        data = {
            "trial": [trial.number for trial in self.history],
            "metric": [trial.metric for trial in self.history],
            "datetime_start": [trial.datetime_start for trial in self.history],
            "datetime_end": [trial.datetime_end for trial in self.history],
            "duration": [trial.duration for trial in self.history],
        }

        for key in self.history[0].hyperparameters.keys():
            data[key] = [trial.hyperparameters[key] for trial in self.history]

        return pd.DataFrame(data)



import os
from typing import NamedTuple, Sequence

import numpy as np
import pandas as pd
from PIL import Image

import torch
from torch import Tensor
from torch.utils.data import Dataset

from ..transformations import Transformation


def scan_classes(path: str) -> list[str]:
    classes: list[str] = []
    for entry in os.scandir(path):
        if entry.is_dir():
            classes.append(entry.name)

    classes.sort()
    return classes


def scan_splits(root: str, allow_alias: bool = True) -> tuple[str, str, str]:
    train_dir_name: str = ""
    validation_dir_name: str = ""
    test_dir_name: str = ""

    splits: list[str] = scan_classes(root)
    for split in splits:
        split_lower = split.lower()
        if "train" in split_lower:
            train_dir_name = split
        elif "val" in split_lower:
            validation_dir_name = split
        elif "test" in split_lower:
            test_dir_name = split

    if not train_dir_name:
        raise ValueError(f"a directory that contains the subword 'train' must be present in {root}")

    if not validation_dir_name and not test_dir_name:
        raise ValueError(f"at least a validation or test split must be present in {root}")

    if (not validation_dir_name or not test_dir_name) and not allow_alias:
        raise ValueError("validation and test splits must both be present unless allow_split_alias=True")

    if not validation_dir_name and test_dir_name:
        validation_dir_name = test_dir_name
    elif not test_dir_name and validation_dir_name:
        test_dir_name = validation_dir_name

    return train_dir_name, validation_dir_name, test_dir_name


class CatastaDataset:
    def __init__(
        self,
        root: str,
        task: str,
        input_name: str | list[str] | None = None,
        output_name: str | list[str] | None = None,
        input_transformations: Sequence[Transformation] | None = None,
        output_transformations: Sequence[Transformation] | None = None,
        grayscale: bool = False,
        allow_split_alias: bool = True,
    ) -> None:
        if task in ("regression", "signal_classification") and (input_name is None or output_name is None):
            raise ValueError("input_name and output_name must be provided for regression and signal_classification tasks")

        self.task: str = task
        self.root: str = root
        if not os.path.isdir(self.root):
            raise ValueError(f"root must be a directory. Found {self.root}")

        input_transformations = [] if input_transformations is None else input_transformations
        output_transformations = [] if output_transformations is None else output_transformations

        splits: tuple[str, str, str] = scan_splits(self.root, allow_alias=allow_split_alias)

        if self.task == "regression":
            self.train = RegressionSubset(
                os.path.join(self.root, splits[0]),
                input_name,  # type: ignore[arg-type]
                output_name,  # type: ignore[arg-type]
                input_transformations,
                output_transformations,
            )
            self.validation = RegressionSubset(
                os.path.join(self.root, splits[1]),
                input_name,  # type: ignore[arg-type]
                output_name,  # type: ignore[arg-type]
                input_transformations,
                output_transformations,
            )
            self.test = RegressionSubset(
                os.path.join(self.root, splits[2]),
                input_name,  # type: ignore[arg-type]
                output_name,  # type: ignore[arg-type]
                input_transformations,
                output_transformations,
            )

        elif self.task == "classification":
            self.train = ClassificationSubset(
                os.path.join(self.root, splits[0]),
                input_transformations,
                grayscale,
            )
            self.validation = ClassificationSubset(
                os.path.join(self.root, splits[1]),
                input_transformations,
                grayscale,
            )
            self.test = ClassificationSubset(
                os.path.join(self.root, splits[2]),
                input_transformations,
                grayscale,
            )

            if len(self.train.classes) == 0:
                raise ValueError(f"no classes found in {self.train.path}")
            if len(self.train.classes) != len(self.validation.classes):
                raise ValueError("number of classes in train and validation splits must be the same")
            if len(self.train.classes) != len(self.test.classes):
                raise ValueError("number of classes in train and test splits must be the same")

        elif self.task == "signal_classification":
            self.train = SignalClassificationSubset(
                os.path.join(self.root, splits[0]),
                input_name,  # type: ignore[arg-type]
                output_name,  # type: ignore[arg-type]
                input_transformations,
            )

            self.validation = SignalClassificationSubset(
                os.path.join(self.root, splits[1]),
                input_name,  # type: ignore[arg-type]
                output_name,  # type: ignore[arg-type]
                input_transformations,
                allowed_labels=self.train.known_labels,
            )

            self.test = SignalClassificationSubset(
                os.path.join(self.root, splits[2]),
                input_name,  # type: ignore[arg-type]
                output_name,  # type: ignore[arg-type]
                input_transformations,
                allowed_labels=self.train.known_labels,
            )

            self.classes = [str(label) for label in self.train.known_labels]
        else:
            raise ValueError(
                "task must be either 'regression', 'classification', or 'signal_classification'. "
                f"Found {task}"
            )


class RegressionSubset(Dataset):
    def __init__(
        self,
        path: str,
        input_name: str | list[str],
        output_name: str | list[str],
        input_transformations: Sequence[Transformation],
        output_transformations: Sequence[Transformation],
    ) -> None:
        self.root: str = path
        self.input_transformations = input_transformations
        self.output_transformations = output_transformations
        self.inputs, self.outputs = self._get_data(input_name, output_name)

        self.inputs = torch.tensor(self.inputs)
        self.outputs = torch.tensor(self.outputs)

    def _get_data(self, input_name: str | list[str], output_name: str | list[str]) -> tuple[np.ndarray, np.ndarray]:
        filenames: list[str] = sorted([name for name in os.listdir(self.root) if name.lower().endswith(".csv")])
        if not filenames:
            raise ValueError(f"Directory {self.root} must contain at least one CSV file")

        inputs: list[np.ndarray] = []
        outputs: list[np.ndarray] = []
        for filename in filenames:
            df = pd.read_csv(os.path.join(self.root, filename))
            input_values = df[input_name].to_numpy()
            output_values = df[output_name].to_numpy()

            for transformation in self.input_transformations:
                input_values = transformation(input_values)
            for transformation in self.output_transformations:
                output_values = transformation(output_values)

            inputs.append(np.asarray(input_values))
            outputs.append(np.asarray(output_values))

        return np.concatenate(inputs), np.concatenate(outputs)

    def __len__(self) -> int:
        return self.inputs.shape[0]

    def __getitem__(self, index: int) -> tuple[Tensor, Tensor]:
        input_tensor: Tensor = self.inputs[index].view(-1)
        if self.outputs.ndim == 1:
            output_tensor = self.outputs[index]
        else:
            output_tensor = self.outputs[index].view(-1)

        return input_tensor, output_tensor


class SignalClassificationSubset(Dataset):
    def __init__(
        self,
        path: str,
        input_name: str | list[str],
        output_name: str | list[str],
        input_transformations: Sequence[Transformation],
        allowed_labels: list[int] | None = None,
    ) -> None:
        self.root: str = path
        self.input_transformations = input_transformations
        inputs, outputs, known_labels = self._get_data(input_name, output_name, allowed_labels)

        self.known_labels = known_labels
        self.inputs = torch.tensor(inputs)
        self.outputs = torch.tensor(outputs, dtype=torch.long)

    def _get_data(
        self,
        input_name: str | list[str],
        output_name: str | list[str],
        allowed_labels: list[int] | None,
    ) -> tuple[np.ndarray, np.ndarray, list[int]]:
        filenames: list[str] = sorted([name for name in os.listdir(self.root) if name.lower().endswith(".csv")])
        if not filenames:
            raise ValueError(f"Directory {self.root} must contain at least one CSV file")

        all_inputs: list[np.ndarray] = []
        all_labels: list[np.ndarray] = []
        for filename in filenames:
            df = pd.read_csv(os.path.join(self.root, filename))
            input_values = df[input_name].to_numpy()
            label_values = df[output_name].to_numpy().reshape(-1)

            for transformation in self.input_transformations:
                input_values = transformation(input_values)

            all_inputs.append(np.asarray(input_values))
            all_labels.append(np.asarray(label_values))

        inputs = np.concatenate(all_inputs)
        labels = np.concatenate(all_labels)
        if labels.ndim != 1:
            labels = labels.reshape(-1)

        if not np.all(np.equal(labels, labels.astype(int))):
            raise ValueError("signal classification labels must be integer class ids")
        label_ids = labels.astype(int)

        if allowed_labels is None:
            known_labels = sorted(np.unique(label_ids).tolist())
        else:
            known_labels = sorted(int(label) for label in allowed_labels)
            unknown_labels = sorted(set(label_ids.tolist()) - set(known_labels))
            if unknown_labels:
                raise ValueError(f"unknown label(s) found in {self.root}: {unknown_labels}")

        label_to_index = {label: index for index, label in enumerate(known_labels)}
        encoded_labels = np.array([label_to_index[label] for label in label_ids], dtype=np.int64)

        return inputs, encoded_labels, known_labels

    def __len__(self) -> int:
        return self.inputs.shape[0]

    def __getitem__(self, index: int) -> tuple[Tensor, Tensor]:
        input_tensor = self.inputs[index].view(-1)
        output_tensor = self.outputs[index]
        return input_tensor, output_tensor


EXTENSIONS = (".jpg", ".jpeg", ".png", ".ppm", ".bmp", ".pgm", ".tif", ".tiff", ".webp")


class Sample(NamedTuple):
    path: str
    label: str


class ClassificationSubset(Dataset):
    def __init__(
        self,
        path: str,
        input_transformations: Sequence[Transformation],
        grayscale: bool,
    ) -> None:
        self.path: str = path
        self.input_transformations = input_transformations
        self.grayscale: bool = grayscale

        self.classes: list[str] = scan_classes(path)
        self.samples: list[Sample] = []
        for class_name in self.classes:
            class_path: str = os.path.join(path, class_name)
            if not os.path.isdir(class_path):
                continue

            for root, _, files in os.walk(class_path):
                for file in files:
                    if not file.lower().endswith(EXTENSIONS):
                        continue

                    self.samples.append(Sample(os.path.join(root, file), class_name))

    def __len__(self) -> int:
        return len(self.samples)

    def __getitem__(self, index: int) -> tuple[Tensor, int]:
        sample = self.samples[index]

        with open(sample.path, "rb") as f:
            image = Image.open(f)
            image = image.convert("RGB" if not self.grayscale else "L")
        sample_array = np.array(image)

        for transformation in self.input_transformations:
            sample_array = transformation(sample_array)

        label: int = self.classes.index(sample.label)
        return torch.tensor(sample_array), label



from .train_info import TrainInfo
from .eval_info import EvalInfo
from .trial_info import TrialInfo
from .optimization_info import OptimizationInfo

__all__ = [
    "TrainInfo",
    "EvalInfo",
    "TrialInfo",
    "OptimizationInfo",
]



import torch
from torch import Tensor
from torch.nn import (
    Conv1d,
    Dropout,
    Flatten,
    GELU,
    Linear,
    MaxPool1d,
    Module,
    ModuleList,
    ReLU,
    Sigmoid,
    Tanh,
)


def get_activation_function(activation: str) -> Module:
    match activation:
        case 'relu':
            return ReLU()
        case 'sigmoid':
            return Sigmoid()
        case 'tanh':
            return Tanh()
        case 'gelu':
            return GELU()
        case _:
            raise ValueError(f'activation function {activation} not supported')


class CNNRegressor(Module):
    def __init__(self, *,
                 n_inputs: int,
                 n_outputs: int,
                 conv_out_channels: list[int],
                 feedforward_dims: list[int] = [],
                 conv_kernel_size: int = 3,
                 conv_stride: int = 1,
                 conv_padding: int = 1,
                 pooling_kernel_size: int = 2,
                 pooling_stride: int = 2,
                 pooling_padding: int = 0,
                 dropout: float = 0.0,
                 activation: str = 'relu',
                 ) -> None:
        super().__init__()
        self.n_outputs = n_outputs

        self.conv_layers = ModuleList()
        for i, out_channels in enumerate(conv_out_channels):
            conv1d = Conv1d(
                in_channels=1 if i == 0 else conv_out_channels[i - 1],
                out_channels=out_channels,
                kernel_size=conv_kernel_size,
                stride=conv_stride,
                padding=conv_padding,
            )
            activation_function = get_activation_function(activation)
            max_pool = MaxPool1d(
                kernel_size=pooling_kernel_size,
                stride=pooling_stride,
                padding=pooling_padding,
            )
            self.conv_layers.append(conv1d)
            self.conv_layers.append(activation_function)
            self.conv_layers.append(max_pool)

        self.flatten = Flatten()

        dummy_input = torch.randn(1, 1, n_inputs)
        for layer in self.conv_layers:
            dummy_input = layer(dummy_input)
        dummy_input = self.flatten(dummy_input)
        prev_neurons = int(dummy_input.shape[1])

        self.dense_layers = ModuleList()
        for i, ff_dim in enumerate(feedforward_dims):
            linear = Linear(
                in_features=prev_neurons if i == 0 else feedforward_dims[i - 1],
                out_features=ff_dim,
            )
            activation_function = get_activation_function(activation)
            dropout_layer = Dropout(p=dropout)
            self.dense_layers.append(linear)
            self.dense_layers.append(activation_function)
            self.dense_layers.append(dropout_layer)

        final_in_features = prev_neurons if len(feedforward_dims) == 0 else feedforward_dims[-1]
        self.dense_layers.append(Linear(final_in_features, n_outputs))

    def forward(self, x: Tensor) -> Tensor:
        if x.ndim == 2:
            x = x.unsqueeze(1)

        for layer in self.conv_layers:
            x = layer(x)

        x = self.flatten(x)

        for layer in self.dense_layers:
            x = layer(x)

        return x.squeeze(-1) if self.n_outputs == 1 else x



from typing import Any

import numpy as np

from ..log import ansi


class EvalInfo:
    def __init__(
        self,
        task: str,
        true_output: np.ndarray,
        predicted_output: Any,
        true_input: np.ndarray | None = None,
        n_classes: int | None = None,
    ) -> None:
        self.task = task
        self.true_input = None if true_input is None else np.asarray(true_input)
        self.true_output = np.asarray(true_output)
        self.predicted_output = np.asarray(predicted_output)

        if self.task == "regression":
            self._calculate_regression_metrics(self.true_output, self.predicted_output)
        elif self.task in ("classification", "signal_classification"):
            self._calculate_classification_metrics(self.true_output, self.predicted_output, n_classes)
        else:
            raise ValueError("Invalid task type")

    @staticmethod
    def _safe_divide(numerator: np.ndarray, denominator: np.ndarray) -> np.ndarray:
        result = np.zeros_like(numerator, dtype=float)
        np.divide(numerator, denominator, out=result, where=denominator != 0)
        return result

    def _calculate_regression_metrics(self, y: np.ndarray, y_pred: np.ndarray) -> None:
        if y.ndim == 1:
            y = y[:, np.newaxis]
        if y_pred.ndim == 1:
            y_pred = y_pred[:, np.newaxis]

        if y.shape != y_pred.shape and y.size == y_pred.size:
            y_pred = y_pred.reshape(y.shape)

        self.mae = np.mean(np.abs(y - y_pred), axis=0)
        self.mse = np.mean((y - y_pred) ** 2, axis=0)
        self.rmse = np.sqrt(self.mse)

        ss_res = np.sum((y - y_pred) ** 2, axis=0)
        ss_tot = np.sum((y - np.mean(y, axis=0)) ** 2, axis=0)
        raw_r2 = self._safe_divide(ss_tot - ss_res, ss_tot)
        self.r2 = np.where(ss_tot == 0, 0.0, raw_r2)

        if self.r2.size == 1:
            self.r2 = float(self.r2.item())  # type: ignore[assignment]

    def _calculate_classification_metrics(self, y: np.ndarray, y_pred: np.ndarray, n_classes: int | None) -> None:
        y_true = np.asarray(y).reshape(-1)
        predictions = np.asarray(y_pred)

        if predictions.ndim > 1:
            y_labels = np.argmax(predictions, axis=1)
        elif predictions.ndim == 1:
            y_labels = predictions
        else:
            raise ValueError("predicted_output must be a 1D label array or a 2D logits/probabilities array")

        y_labels = np.asarray(y_labels).reshape(-1)
        if y_true.shape[0] != y_labels.shape[0]:
            raise ValueError("true_output and predicted_output must contain the same number of samples")

        if not np.all(np.equal(y_true, y_true.astype(int))):
            raise ValueError("true labels must be integer class ids")
        if not np.all(np.equal(y_labels, y_labels.astype(int))):
            raise ValueError("predicted labels must be integer class ids")

        y_true = y_true.astype(int)
        y_labels = y_labels.astype(int)

        if n_classes is None:
            n_classes = int(max(np.max(y_true), np.max(y_labels))) + 1 if y_true.size else 0
        if n_classes <= 0:
            raise ValueError("n_classes must be a positive integer")

        self.confusion_matrix = self._compute_confusion_matrix(y_true, y_labels, n_classes)

        self.tp = np.diag(self.confusion_matrix).astype(float)
        self.fp = (np.sum(self.confusion_matrix, axis=0) - self.tp).astype(float)
        self.fn = (np.sum(self.confusion_matrix, axis=1) - self.tp).astype(float)
        self.tn = (np.sum(self.confusion_matrix) - (self.tp + self.fp + self.fn)).astype(float)

        total = float(np.sum(self.confusion_matrix))
        self.accuracy = float(np.sum(self.tp) / total) if total > 0 else 0.0

        precision_per_class = self._safe_divide(self.tp, self.tp + self.fp)
        recall_per_class = self._safe_divide(self.tp, self.tp + self.fn)
        specificity_per_class = self._safe_divide(self.tn, self.tn + self.fp)
        f1_per_class = self._safe_divide(2 * precision_per_class * recall_per_class, precision_per_class + recall_per_class)

        self.precision = float(np.mean(precision_per_class))
        self.recall = float(np.mean(recall_per_class))
        self.sensitivity = self.recall
        self.specificity = float(np.mean(specificity_per_class))
        self.f1_score = float(np.mean(f1_per_class))

    def _compute_confusion_matrix(self, y: np.ndarray, y_pred: np.ndarray, classes: int) -> np.ndarray:
        if np.any((y < 0) | (y >= classes)):
            raise ValueError(f"true label out of range for n_classes={classes}")
        if np.any((y_pred < 0) | (y_pred >= classes)):
            raise ValueError(f"predicted label out of range for n_classes={classes}")

        matrix = np.zeros((classes, classes), dtype=int)
        for true, pred in zip(y, y_pred):
            matrix[int(true), int(pred)] += 1

        return matrix

    def _regression_repr(self) -> str:
        return (
            f"\n{ansi.BOLD}{ansi.BLUE}-> regression metrics{ansi.RESET}\n"
            f"   |> MAE:  {np.array2string(np.asarray(self.mae), precision=4, floatmode='fixed')}\n"
            f"   |> MSE:  {np.array2string(np.asarray(self.mse), precision=4, floatmode='fixed')}\n"
            f"   |> RMSE: {np.array2string(np.asarray(self.rmse), precision=4, floatmode='fixed')}\n"
            f"   |> R2:   {np.array2string(np.asarray(self.r2), precision=4, floatmode='fixed')}"
        )

    def _classification_repr(self) -> str:
        return (
            f"\n{ansi.BOLD}{ansi.BLUE}-> classification metrics{ansi.RESET}\n"
            f"   |> accuracy:    {self.accuracy:.4f}\n"
            f"   |> precision:   {self.precision:.4f}\n"
            f"   |> recall:      {self.recall:.4f}\n"
            f"   |> sensitivity: {self.sensitivity:.4f}\n"
            f"   |> specificity: {self.specificity:.4f}\n"
            f"   |> f1 score:    {self.f1_score:.4f}"
        )

    def __repr__(self) -> str:
        if self.task == "regression":
            return self._regression_repr()
        if self.task in ("classification", "signal_classification"):
            return self._classification_repr()
        raise ValueError("Invalid task type")



import numpy as np

from .transformation import Transformation


class Normalization(Transformation):
    def __init__(self, norm_tech: str) -> None:
        """
        Initialize the normalization technique

        Parameters
        ----------
        norm_tech : str
            Normalization technique to be applied. The following techniques are supported:
            - "arctan"
            - "boxcox"
            - "decimal"
            - "logistic"
            - "mad"
            - "max"
            - "maxabs"
            - "maxmin"
            - "mean"
            - "median"
            - "min"
            - "minabs"
            - "minmax"
            - "power"
            - "quantile"
            - "robust"
            - "softmax"
            - "std"
            - "tanh"
            - "var"
            - "yeojohnson"
            - "zscore"
        """
        self.norm_tech: str = norm_tech

    @staticmethod
    def _safe_divide(numerator: np.ndarray, denominator: float, eps: float = 1e-12) -> np.ndarray:
        if abs(float(denominator)) <= eps:
            return np.zeros_like(numerator, dtype=float)
        return numerator / denominator

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Apply normalization technique to the input data

        Parameters
        ----------
        x : np.ndarray
            Input data to be normalized

        Returns
        -------
        np.ndarray
            Normalized data
        """
        x = np.asarray(x, dtype=float)

        match self.norm_tech:
            case "arctan":
                return np.arctan(x)
            case "boxcox":
                return np.sign(x) * np.power(np.abs(x), 1 / 2)
            case "decimal":
                max_abs = float(np.max(np.abs(x)))
                if max_abs <= 0.0:
                    return np.zeros_like(x, dtype=float)
                scale = 10 ** np.ceil(np.log10(max_abs))
                return self._safe_divide(x, float(scale))
            case "logistic":
                return 1 / (1 + np.exp(-x))
            case "mad":
                mad = float(np.median(np.abs(x - np.median(x))))
                return self._safe_divide(x, mad)
            case "max":
                return self._safe_divide(x, float(np.max(x)))
            case "maxabs":
                return self._safe_divide(x, float(np.max(np.abs(x))))
            case "maxmin":
                return self._safe_divide(x, float(np.max(x) - np.min(x)))
            case "mean":
                return self._safe_divide(x, float(np.mean(x)))
            case "median":
                return self._safe_divide(x, float(np.median(x)))
            case "min":
                return self._safe_divide(x, float(np.min(x)))
            case "minabs":
                return self._safe_divide(x, float(np.min(np.abs(x))))
            case "minmax":
                min_value = float(np.min(x))
                range_value = float(np.max(x) - min_value)
                return self._safe_divide(x - min_value, range_value)
            case "power":
                return np.sign(x) * np.power(np.abs(x), 1 / 2)
            case "quantile":
                quantile = float(np.quantile(x, 0.95))
                return self._safe_divide(x, quantile)
            case "robust":
                iqr = float(np.percentile(x, 75) - np.percentile(x, 25))
                return self._safe_divide(x, iqr)
            case "softmax":
                shifted = x - np.max(x)
                exp_shifted = np.exp(shifted)
                return self._safe_divide(exp_shifted, float(np.sum(exp_shifted)))
            case "std":
                return self._safe_divide(x, float(np.std(x)))
            case "tanh":
                return np.tanh(x)
            case "var":
                return self._safe_divide(x, float(np.var(x)))
            case "yeojohnson":
                return np.sign(x) * np.power(np.abs(x), 1 / 2)
            case "zscore":
                return self._safe_divide(x - np.mean(x), float(np.std(x)))
            case _:
                raise ValueError("Unknown normalization technique")



from __future__ import annotations

from copy import deepcopy
from typing import Literal

from torch.nn import Module

EarlyStoppingMode = Literal["none", "trend", "plateau"]
EarlyStoppingPolicy = tuple[str, float | int, float | int]


class ModelStateManager:
    def __init__(self, mode: str | EarlyStoppingPolicy | None = "none") -> None:
        self.mode, self.alpha, self.patience, self.min_delta = self._parse_mode(mode)

        self.best_loss: float = float("inf")
        self.best_model_state: dict = {}

        self.stop: bool = False
        self.prev_loss: float | None = None
        self.derivative: float = 0.0
        self.bad_epochs: int = 0
        self.trend_positive_epochs: int = 0

    def _parse_mode(
        self,
        mode: str | EarlyStoppingPolicy | None,
    ) -> tuple[EarlyStoppingMode, float, int, float]:
        if mode is None or mode == "none":
            return "none", 0.0, 0, 0.0

        if not isinstance(mode, tuple) or len(mode) != 3 or not isinstance(mode[0], str):
            raise ValueError(
                "early_stopping must be one of: None, 'none', ('trend', alpha, patience), ('plateau', patience, min_delta)"
            )

        tag = mode[0].lower()
        if tag == "trend":
            alpha = float(mode[1])
            patience = int(mode[2])

            if not (0 <= alpha < 1):
                raise ValueError("trend alpha must be in [0, 1)")
            if patience < 1:
                raise ValueError("trend patience must be >= 1")

            return "trend", alpha, patience, 0.0

        if tag == "plateau":
            patience = int(mode[1])
            min_delta = float(mode[2])

            if patience < 1:
                raise ValueError("plateau patience must be >= 1")
            if min_delta < 0:
                raise ValueError("plateau min_delta must be >= 0")

            return "plateau", 0.0, patience, min_delta

        raise ValueError(
            "early_stopping must be one of: None, 'none', ('trend', alpha, patience), ('plateau', patience, min_delta)"
        )

    def __call__(self, model: Module, loss: float) -> None:
        previous_best_loss = self.best_loss

        if not self.best_model_state:
            self.best_model_state = deepcopy(model.state_dict())

        if loss < self.best_loss:
            self.best_loss = loss
            self.best_model_state = deepcopy(model.state_dict())

        if self.mode == "none":
            return

        if self.mode == "trend":
            delta = 0.0 if self.prev_loss is None else loss - self.prev_loss
            self.derivative = self.alpha * self.derivative + (1 - self.alpha) * delta
            self.prev_loss = loss

            if self.derivative > 0:
                self.trend_positive_epochs += 1
            else:
                self.trend_positive_epochs = 0

            if self.trend_positive_epochs >= self.patience:
                self.stop = True
            return

        if self.mode == "plateau":
            if loss < (previous_best_loss - self.min_delta):
                self.bad_epochs = 0
            else:
                self.bad_epochs += 1

            if self.bad_epochs >= self.patience:
                self.stop = True
            return

    def load_best_model_state(self, model: Module) -> None:
        if self.best_model_state:
            model.load_state_dict(self.best_model_state)
